#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "general.h"

/*

imelody2wav - Copyright 2001 Michael Kohn (mike@naken.cc)
This falls under the Kohnian license.  Please read
it at http://nakentone.naken.cc/

This program is NOT opensourced.  You may not use any part
of this program for your own software.

*/

int pushback=-1;

struct note_t
{
  int sample_rate;
  int bytes;
  int scale;
  int tone;
  int length;
  int modifier;
  unsigned int samples_per_beat;
};

int write_long(FILE *out, int n)
{
  putc((n&255),out);
  putc(((n>>8)&255),out);
  putc(((n>>16)&255),out);
  putc(((n>>24)&255),out);

  return 0;
}

int write_word(FILE *out, int n)
{
  putc((n&255),out);
  putc(((n>>8)&255),out);

  return 0;
}

void write_note(FILE *out, struct note_t *note)
{
unsigned int samples_per_note;
unsigned int samples_per_wave;
float waveform;
float waveform_slope;
float actual_freq;
/*              P  C        C#       C        D#       E        F */
float freqs[]={ 0, 261.625, 227.175, 293.675, 311.125, 329.625, 349.225, 
                370, 392, 415.3, 440, 466.15, 246.95 };
/*              F#   G    G#     A    A#      B */
int t;
unsigned short int sample16;
unsigned char sample8;

  samples_per_note=note->samples_per_beat*(4/(note->length+1));

  if (note->tone==0)
  {
    for (t=0; t<samples_per_note; t++)
    {
      if (note->bytes==1)
      { putc(0,out); }
        else
      { putc(0,out); putc(0,out); }
    }
  }
    else
  {
    if (note->tone>12) note->tone=1;
    actual_freq=freqs[note->tone];
    actual_freq=actual_freq*(float)(note->scale);
    samples_per_wave=note->sample_rate/actual_freq;

    if (note->bytes==1)
    {
      waveform_slope=255/(float)samples_per_wave;
      waveform=127;

      for (t=0; t<samples_per_note; t++)
      {
        sample8=(char)waveform;
        putc(sample8,out);
        waveform=waveform-waveform_slope;
        if (waveform<-127) waveform=127;
      }

      for (t=0; t<20; t++)
      {
        putc(0,out);
      }
    }
      else
    if (note->bytes==2)
    {
      waveform_slope=65535/(float)samples_per_wave;
      waveform=32767;

      for (t=0; t<samples_per_note; t++)
      {
        sample16=waveform;
        write_word(out,sample16);
        waveform=waveform-waveform_slope;
        if (waveform<-32767) waveform=32767;
      }

      for (t=0; t<20; t++)
      {
        write_word(out,0);
      }
    } 
  }
}

int main (int argc, char *argv[])
{
FILE *in,*out;
char filename[1024];
char outname[1024];
char songname[16];
char song[1024];
int t,l,ptr,ch;
int b,scale;
int length,tone,octave,modifier;
int tone_modifier,ats;
int default_octave,marker;
struct note_t note;
unsigned int data_length;

  scale=-1;
  b=140;
  ats=1;
  default_octave=1;
  note.sample_rate=44100;
  note.bytes=2;

  strcpy(outname,"imelody.wav");

  if (argc<2)
  {
    printf("\nimelody2wav "VERSION" - Convert Ericsson imelody files to .wav.\n");
    printf(COPYRIGHT);
    printf("Usage: imelody2wav [ options ] <input filename>\n");
    printf("       -o <output filename [default: imelody.wav]>\n");
    printf("       -t <tempo [default: %d]>\n",b);
    printf("       -a [force imelody2wav to ignore @'s in song]>\n");
    printf("       -f <sampling frequency [default: 44100]>\n");
    printf("       -b <bits 8 or 16 [default: 16]>\n\n");
    exit(0);
  }

  for (t=1; t<argc; t++)
  {
    if (strcmp(argv[t],"-a")==0)
    { ats=0; }
      else
    if (strcmp(argv[t],"-t")==0)
    { 
      b=atoi(argv[++t]);
    }
      else
    if (strcmp(argv[t],"-f")==0)
    { 
      note.sample_rate=atoi(argv[++t]);
    }
      else
    if (strcmp(argv[t],"-b")==0)
    { 
      note.bytes=atoi(argv[++t]);
      if (note.bytes==16)
      { note.bytes=2; }
        else
      { note.bytes=1; }
    }
      else
    if (strcmp(argv[t],"-o")==0)
    { 
      strcpy(outname,argv[++t]);
    }
      else
    { strncpy(filename,argv[t],1023); }
  }

  in=fopen(filename,"r");
  if (in==0)
  {
    printf("Could not open file for reading: %s\n",filename);
    exit(0);
  }

  out=fopen(outname,"w");
  if (out==0)
  {
    printf("Could not open file for writing: %s\n",outname);
    fclose(in);
    exit(0);
  }

  /* Write RIFF header */

  fprintf(out,"RIFF");
  write_long(out,4);
  fprintf(out,"WAVE");

  /* Write fmt header */

  fprintf(out,"fmt ");
  write_long(out,16);
  write_word(out,1);
  write_word(out,1);
  write_long(out,note.sample_rate);
  write_long(out,note.sample_rate*note.bytes);
  if (note.bytes==1)
  { write_word(out,1); }
    else
  if (note.bytes==2)
  { write_word(out,2); }
  write_word(out,note.bytes*8);

  /* Write the start of the data header */

  data_length=0;
  fprintf(out,"data");
  marker=ftell(out);
  write_long(out,data_length);

  note.samples_per_beat=note.sample_rate*60/b;

  ptr=0;
  while ((ch=getc(in))!=EOF)
  { song[ptr++]=ch; }

  song[ptr]=0;
  fclose(in);

  l=0;
  while (song[l]!=':' && l<15)
  {
    songname[l]=song[l++];
  }

  songname[l]=0;

  printf("\nimelody2wav "VERSION" - Convert Ericsson imelody files to .wav.\n");
  printf(COPYRIGHT);

  printf("\n       Song Name: %s\n",songname);

  printf("Beats Per Minute: %d\n",b);

  /* t=get_tempo(b); */

  l=0;
  while (song[l]!=':' && l<ptr) l++;
  l++;
 
  for (t=l; t<ptr; t++)
  {
    if (song[t]=='r') song[t]='p';
    if (song[t]=='R') song[t]='P';
  }
 
  while (l<ptr)
  {
    if (song[l]==' ' || song[l]=='\t' || song[l]=='\r' || song[l]=='\n')
    { l++; continue; }

    if (song[l]=='@' && ats==0)
    { l++; continue; }

    if (song[l]=='*')
    {
      l++;
      default_octave=(song[l++]-'0')-3;
    }

    if (song[l]=='+' && song[l+1]=='+') { octave=2+default_octave; l=l+2; }
      else
    if (song[l]=='+') { octave=1+default_octave; l++; }
      else
    { octave=default_octave; }

    tone_modifier=0;

    if (song[l]=='#')
    {
      tone_modifier=tone_modifier+1;
      l++;
    }

    if (song[l]=='&' || song[l]=='@')
    {
      tone_modifier=tone_modifier-1;
      l++;
    }

    if (song[l]=='(' && song[l+1]=='b' && song[l+2]==')')
    {
      tone_modifier=tone_modifier-1;
      l=l+3;
    }

    if (tolower(song[l])=='a') tone=10; 
      else
    if (tolower(song[l])=='h' || tolower(song[l])=='b') tone=12;
      else
    if (tolower(song[l])=='c') tone=1;
      else
    if (tolower(song[l])=='d') tone=3;
      else
    if (tolower(song[l])=='e') tone=5;
      else
    if (tolower(song[l])=='f') tone=6;
      else
    if (tolower(song[l])=='g') tone=8;
      else
    if (tolower(song[l])=='p') tone=0;

    if (song[l]>='a' && song[l]<='z') length=3;
      else
    if (song[l]>='A' && song[l]<='Z') length=2;

    if (song[l]=='p')
    {
      if (song[l+1]=='p' && song[l+2]=='p' && song[l+3]=='p')
      { length=1; l=l+4; }
        else
      if (song[l+1]=='p')
      { length=2; l=l+2; }
        else
      { l++; }
    }
      else
    { l++; }

    if (song[l]=='.')
    {
      modifier=1;
      l++;
    }

    if (song[l]>='0' && song[l]<='3')
    {
      if (song[l]=='0') 
      { modifier=1; length=2; } 
        else
      if (song[l]=='1') 
      { modifier=0; length=2; } 
        else
      if (song[l]=='2') 
      { modifier=1; length=3; } 
        else
      if (song[l]=='3') 
      { modifier=0; length=3; } 

      l++;
    }

    tone=tone+tone_modifier;

    if (scale!=octave)
    {
      /* if (octave==5) warning++; */
      scale=octave;
    }

    modifier=0;

    if (song[l]=='.')
    {
      modifier=1;
      l++;
    }

/*
    push(1,3);
    push(tone,4);
    push(length,3);
    push(modifier,2);
*/

    note.tone=tone;
    note.length=length;
    note.modifier=modifier;
    note.scale=scale;

    write_note(out,&note);
  }

  data_length=ftell(out)-marker-4;
  fseek(out,marker,0);
  write_long(out,data_length);

  return (0);
}



